const { app, dialog } = require('electron');
const child_process = require('child_process');
const fs = require("fs");
const path = require("path");
const portConfig = require("./port-config.js");
const { Writable } = require('stream');
const http = require("http");

let _hasError = false;
let _isAlreadyOpen = false;
let _errorMessage = "";

function launch() {
    let serverProcessId = getServerProcessId();
    if (serverProcessId >= 0) {
        // to do: when server process exits, app.quit() 

        // could try to keep open a websocket and poll using find-process

    }
    else {
        // consider getting this to work properly, to help ensure consistent behavior
        //portConfig.reset();

        let dir = path.dirname(process.execPath);
        let binDir = path.join(dir, "bin");

        let cwd = process.cwd();
        let cmd = path.join(binDir, "Blockpad");
        if (process.platform === "win32")
            cmd += ".exe";
        let childProcess = child_process.spawn(cmd, ["--gui-running", process.pid], {
            cwd: binDir,
            env: process.env
        });

        let childProcessOutput = createMemoryStream();
        childProcess.stdout.pipe(childProcessOutput);
        childProcess.stderr.pipe(childProcessOutput);
    
        childProcess.on("exit", async () => {
            //console.log("Background process exited");
            if (_isUIReady) {
                //console.log("UI had been loaded - quit")
                app.quit();
            }
            else if (await isReady()) {
                //console.log("Another Blockpad background process is open")
                _hasError = true;
                _isAlreadyOpen = true;
            }
            else {
                //console.log("Unexpected error, showing error page")
                _hasError = true;
                _errorMessage = readMemoryStream(childProcessOutput);
            }
        });

        childProcess.on("error", error => {
            _hasError = true;
            _errorMessage = "Error starting background process\n\n" + error.message;
        });

        childProcess.on("spawn", () => {
            // success
        });
    }
}

function getServerProcessId() {
    for (let i = 0; i<process.argv.length; i++){
        let arg = process.argv[i];
        if (arg === "--server-running") {
            if (process.argv.length > i + 1) {
                let valueArg = process.argv[i + 1];
                if (valueArg) {
                    return Number(valueArg);
                }
            }
            else {
                return 0;
            }
        }
    }
    return -1;
}

async function isReady() {
    let statusCode = 0;
    let port = 0;
    let responseData;

    await new Promise(resolve => {
        try {
            port = portConfig.getAppPort();
            if (!portConfig.isValid(port)) {
                win.once("ready-to-show", () => {
                    win.show();
                    dialog.showMessageBox({
                        type: 'info',
                        title: 'Blockpad',
                        message: "Unable to find local port to start Blockpad. It may help to restart your computer and try again. If this problem persists, please contact support.",
                        buttons: ['OK']
                    });
                    app.quit();
                });
                return;
            }
            url = `http://localhost:${port}/check`;
            //console.log(url)
            responseData = "";
            let request = http.get(url, response => {
                statusCode = response.statusCode;
                //console.log("statusCode");

                if (statusCode >= 200 && statusCode < 300) {
                    response.setEncoding('utf-8');

                    response.on('data', (chunk) => {
                        responseData += chunk;
                    });

                    response.on('end', () => {
                        //console.log("data: " + responseData)
                        resolve();
                    });
                }
                else {
                    //console.log("response: " + statusCode)
                    resolve();
                }
            });
            request.on("error", () => {
                //console.log("error")
                resolve();
            });
        }
        catch (ex) {
            //console.log("Exception: " + ex)
            resolve();
        }
    });

    //console.log(responseData)
    if (responseData === "Welcome to Blockpad!") {
        portConfig.confirmPort(port);
        return true;
    }
    else {
        return false;
    }
}

function createMemoryStream() {
    let stream = new Writable({
        write(chunk, encoding, callback) {
            if (encoding === 'buffer') {
                this.buffer = Buffer.concat([this.buffer, chunk]);
            } else {
                this.buffer = Buffer.concat([this.buffer, Buffer.from(chunk, encoding)]);
            }
            callback();
        }
    });
    stream.buffer = Buffer.alloc(0);
    return stream;
}

function readMemoryStream(stream) {
    let str = stream.buffer.toString("utf8");
    return str;
}

function hasError() {
    return _hasError;
}

function isAlreadyOpen() {
    return _isAlreadyOpen;
}

function getErrorMessage() {
    return _errorMessage;
}

let _isUIReady = false;

function markUIReady() {
    _isUIReady = true;
}

exports.launch = launch;
exports.isReady = isReady;
exports.hasError = hasError;
exports.getErrorMessage = getErrorMessage;
exports.markUIReady = markUIReady;
exports.isAlreadyOpen = isAlreadyOpen;