﻿const { app, BrowserWindow, Menu, ipcMain, systemPreferences, dialog } = require('electron');
const dialogs = require("./dialogs.js");
const portConfig = require("./port-config.js");
const localServer = require("./local-server.js");

const isMac = process.platform === "darwin";
const isWindows = process.platform === "win32";

let queuedAction;

function createWindow() {
    let options = {
        width: 1200,
        height: 800,
        icon: "icon.ico",
        title: "Blockpad",
        backgroundColor: "#fff",
        show: false,
        webPreferences: {
            nodeIntegration: true,
            partition: 'nopersist', // No persistent session, disables cache
            //devTools: false,
            enableRemoteModule: true,
            contextIsolation: false
        }
    };

    let urlSuffix = "";

    if (isMac) {
        options.titleBarStyle = "hiddenInset";
        urlSuffix += "&mac=true";
    }
    else if (isWindows) {
        options.titleBarStyle = "hidden",
            options.titleBarOverlay =
            {
                color: "#222222",
                symbolColor: "#ffffff"
            }

    }

    const win = new BrowserWindow(options);

    let getUrl = port => `http://localhost:${port}/main-window?id=${win.id}` + urlSuffix;

    win.setMenu(null);

    win.on("focus", () => {
        activeWindow = win;
        win.webContents.send("focus");
    });

    win.on("shown", () => {
        win.webContents.send("shown");
    });

    if (isMac) {
        applyMacMenu(win, getDefaultMacMenu(false));
    }
    else {
        win.maximize();
    }

    //win.webContents.on("did-fail-load", () => {
    //    setTimeout(() => win.loadURL(url), 500);
    //});

    win.webContents.once('did-finish-load', () => {
        win.show();

        try {
            goToUrlWhenReady(win, getUrl);
        }
        catch (ex) {
            dialog.showMessageBox({
                type: 'info',
                title: 'Blockpad',
                message: ex.message,
                buttons: ['OK']
            });
            win.close();
        }
    });

    win.on("close", async e => {
        let url = win.webContents.getURL();
        if (!url.startsWith("file:///") && !url.startsWith("data:")) {
            e.preventDefault();

            win.webContents.send("try-close-window", win.id.toString());
        }
    });

    //win.loadURL(url);
    win.loadFile("waiting.html");
}

let firstTime = true;

async function goToUrlWhenReady(win, getUrl) {
    await new Promise(resolve => setTimeout(resolve, 1));

    if (firstTime) {
        firstTime = false;
        try {
            await win.webContents.session.clearCache();
        }
        catch (ex) {
            console.log("Error clearing cache: " + ex)
        }
    }

    while (true) {
        console.log("attempting...")
        let ready = await localServer.isReady();

        await new Promise(resolve => setTimeout(resolve, 100));

        if (localServer.hasError()) {
            if (localServer.isAlreadyOpen()) {
                dialog.showMessageBox({
                    type: 'info',
                    title: 'Blockpad',
                    message: "Blockpad is already open",
                    buttons: ['OK']
                });
                win.close();
                app.quit();
                return;
            }
            win.loadURL("data:text/html;charset=UTF-8," + encodeURIComponent(getErrorPageHtml()));
            //win.webContents.openDevTools();
            applyMacMenu(win, null);
            return;
        }
        else if (ready) {
            break;
        }
    }

    localServer.markUIReady();

    let url = getUrl(portConfig.getAppPort());
    win.loadURL(url);
    setTimeout(() => {
        if (queuedAction) {
            queuedAction(win);
            queuedAction = null;
        }
    }, 1000);
}

function newTab(path) {
    whenReady(win => {
        win.webContents.send("new-tab", win.id.toString(), path);
    });
}

function whenReady(action) {
    let win = getActiveWindow();
    if (win) {
        action(win);
    }
    else {
        let prevAction = queuedAction;
        queuedAction = win => {
            if (prevAction)
                prevAction(win);
            action(win);
        }
    }
}

function getActiveWindow() {
    return BrowserWindow.getFocusedWindow() ?? BrowserWindow.getAllWindows()[0];
}

ipcMain.handle("close-main-window", (event, browserWindowId) => {
    //console.log("Handling close-main-window, " + browserWindowId.toString());
    let window = BrowserWindow.fromId(Number(browserWindowId));
    //console.log(`window = ${window}, event = ${JSON.stringify(Object.keys(event))}, sender = ${event.sender}`);

    let closeDialog = closeDialogs[window.id];
    if (closeDialog) {
        //console.log("Closing close dialog");
        closeDialog.destroy();
        delete closeDialogs[window.id];
    }

    window.destroy();
});

ipcMain.handle("show-close-dialog", (event, browserWindowId, tabIds) => {
    let window = BrowserWindow.fromId(Number(browserWindowId));
    showCloseDialog(window, tabIds);
});

ipcMain.handle("set-mac-menu", (event, browserWindowId, menu) => {
    let window = BrowserWindow.fromId(Number(browserWindowId));
    applyMacMenu(window, menu);
});

ipcMain.handle("set-tab-title", (event, browserWindowId, title) => {
    let window = BrowserWindow.fromId(Number(browserWindowId));
    if (title)
        window.setTitle(`${title} - Blockpad`);
    else
        window.setTitle("Blockpad");
});

ipcMain.handle("double-click-titlebar", (event, browserWindowId) => {
    let win = BrowserWindow.fromId(Number(browserWindowId));
    const doubleClickAction = systemPreferences.getUserDefault('AppleActionOnDoubleClick', 'string');
    if (doubleClickAction === 'Minimize') {
        win.minimize();
    }
    else if (doubleClickAction === 'Maximize') {
        if (!win.isMaximized()) {
            win.maximize();
        } else {
            win.unmaximize();
        }
    }
});

ipcMain.handle("show-dev-tools", (event, browserWindowId) => {
    let win = BrowserWindow.fromId(Number(browserWindowId));
    win.openDevTools();
});

ipcMain.handle("activate-window", (event, browserWindowId) => {
    let win = BrowserWindow.fromId(Number(browserWindowId));
    win.focus();
});

ipcMain.handle("new-main-window", () => {
    createWindow();
});

function applyMacMenu(window, template) {
    if (template) {
        for (let item of template)
            applyClickEvents(item);

        let menu = Menu.buildFromTemplate(template);

        Menu.setApplicationMenu(menu);
    }
    else {
        let menu = Menu.buildFromTemplate(getDefaultMacMenu(true));

        Menu.setApplicationMenu(menu);
    }
}

function getDefaultMacMenu(includeBasics) {
    let menu = [
        {
            label: "Blockpad",
            submenu: [
                { role: "services" },
                { type: "separator" },
                { role: "hide" },
                { role: "hideothers" },
                { role: "unhide" },
                { type: "separator" },
                { role: "quit" },
            ]
        }
    ];
    if (includeBasics) {
        menu.push({
            label: "File",
            submenu: [
                {
                    label: "New Window",
                    itemPath: "Library.Tools.File.NewWindowTool"
                },
                {
                    label: "Close",
                    itemPath: "Library.Tools.File.CloseTool"
                }
            ]
        });
        menu.push({
            label: "Edit",
            submenu: [
                {
                    label: "Cut",
                    role: "cut",
                    accelerator: "CommandOrControl+X"
                },
                {
                    label: "Cut",
                    role: "copy",
                    accelerator: "CommandOrControl+C"
                },
                {
                    label: "Cut",
                    role: "paste",
                    accelerator: "CommandOrControl+V"
                }
            ]
        });
        menu.push({
            label: "Help",
            submenu: [
                {
                    label: "About",
                    itemPath: "Library.Tools.Help.AboutTool"
                }
            ]
        });
    }
    return menu;
}

function applyClickEvents(menuTemplate) {
    if (menuTemplate.submenu) {
        for (let item of menuTemplate.submenu) {
            applyClickEvents(item);
        }
    }
    else if (!menuTemplate.click) {
        let path = menuTemplate.itemPath;
        if (path === "Library.Tools.Edit.CopyTool" && focusedOnTextBox()) {
            menuTemplate.role = "copy";
            delete menuTemplate.enabled;
        }
        else if (path === "Library.Tools.Edit.CutTool" && focusedOnTextBox()) {
            menuTemplate.role = "cut";
            delete menuTemplate.enabled;
        }
        else if (path === "Library.Tools.Edit.PasteTool" && focusedOnTextBox()) {
            menuTemplate.role = "paste";
            delete menuTemplate.enabled;
        }
        else if (path) {
            menuTemplate.click = menuItemClick;
        }
    }
}

function focusedOnTextBox() {
    return true;
}

function menuItemClick(menuItem, browserWindow, event) {
    let path = menuItem.itemPath;
    let id = -1;
    if (browserWindow)
        id = browserWindow.id;
    browserWindow.webContents.send("run-tool", id.toString(), path);
}

let closeDialogs = {};

function showCloseDialog(win, tabIds) {
    if (closeDialogs[win.id])
        return;

    port = portConfig.getAppPort();

    let promptWindow = new BrowserWindow({
        parent: win,
        modal: true,
        width: 600,
        height: 300,
        icon: "icon.ico",
        title: "Blockpad",
        webPreferences: {
            nodeIntegration: true,
            //devTools: false,
            partition: 'nopersist', // No persistent session, disables cache
            enableRemoteModule: true,
            contextIsolation: false
        }
    });

    promptWindow.once("ready-to-show", () => {
        promptWindow.show();
    });

    promptWindow.on("closed", () => {
        delete closeDialogs[win.id];
    });

    promptWindow.setMenu(null);

    closeDialogs[win.id] = promptWindow;

    let url = `http://localhost:${port}/confirm-close?id=${win.id}`;

    if (tabIds) {
        for (let id of tabIds)
            url += `&tab=${id}`;
    }

    promptWindow.loadURL(url);
}

function getErrorPageHtml() {
    let errorMessage = localServer.getErrorMessage().replace("Welcome to Blockpad!", "");
    let maxLength = 25000;
    if (errorMessage.length > maxLength)
        errorMessage.substring(0, maxLength) + "\n...";
    let encodedErrorMessage = errorMessage.replace("&", "&amp").replace("<", "&lt;").replace(">", "&gt;");
    return `<html><body style="margin: 0; overflow: hidden; background: #eee; box-sizing: border-box; height: 100%; font-family: system-ui, sans-serif; display: flex; flex-direction: column; justify-content: stretch; align-items: stretch; padding: 30px; -webkit-app-region: drag; "><div style="margin: 30px 10px 10px 10px;">There was an error starting Blockpad. See detailed info below. Please contact support to resolve the issue.</div><div style="margin: 10px; background: #fff; border: solid 1px #999; border-radius: 8px; padding: 8px; overflow: scroll; user-select: auto; white-space: pre; flex-grow: 1; -webkit-app-region: no-drag;">${encodedErrorMessage}</div></body></html>`;
}

module.exports = { createWindow, newTab };